/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2003, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// evaluate reverse polish notation mathematic operations.

#include <iostream>
#include <string>

using namespace std;


typedef enum {
	kOperand,
	kAdd,
	kSubtract,
	kMultiply,
	kDivide
} RPNType;

class RPNStep {
public:
	RPNStep(void);
	RPNStep(RPNType type, double operand=0.0); // operand is irrelevant unless type==kOperand
	RPNStep(const RPNStep& step);

	~RPNStep();

	const RPNStep& operator = (const RPNStep& step);

	// evaluate this, returns false if error like stack underflow for example
	bool Execute(double *bottom, double **pstack);

	RPNType GetType(void) const { return mType; }
	double GetOperand(void) const { return mOperand; }

protected:
	RPNType mType;
	double mOperand; // irrelevant unless mType==kOperand
};


RPNStep::RPNStep(void)
{
}

RPNStep::RPNStep(RPNType type, double operand)
{
	mType = type;
	mOperand = operand;
}

RPNStep::RPNStep(const RPNStep& step)
{
	mType = step.mType;
	mOperand = step.mOperand;
}

RPNStep::~RPNStep()
{
}

const RPNStep& RPNStep::operator = (const RPNStep& step)
{
	mType = step.mType;
	mOperand = step.mOperand;
	return *this;
}

bool CheckFor2(double *bottom, double *stack)
{
	stack -= 2;
	return ( stack >= bottom );
}

// evaluate this 
bool RPNStep::Execute(double *bottom, double **pstack)
{
	double top, bot, *stack = *pstack;

	switch(mType)
	{
	case kOperand:
		*stack++ = mOperand;
		break;

	case kAdd:
		if( !CheckFor2(bottom,stack) )
			return false;
		top = *(--stack);
		bot = *(--stack);
		*stack++ = bot + top;
		break;

	case kSubtract:
		if( !CheckFor2(bottom,stack) )
			return false;
		top = *(--stack);
		bot = *(--stack);
		*stack++ = bot - top;
		break;

	case kMultiply:
		if( !CheckFor2(bottom,stack) )
			return false;
		top = *(--stack);
		bot = *(--stack);
		*stack++ = bot * top;
		break;

	case kDivide:
		if( !CheckFor2(bottom,stack) )
			return false;
		top = *(--stack);
		bot = *(--stack);
		if( !top )
			return false; // divide by zero
		*stack++ = bot / top;
		break;

	}

	*pstack = stack;
	return true;
}

// display whatever this is
ostream& operator << ( ostream& os, const RPNStep& step )
{
	switch(step.GetType())
	{
	case kOperand:
		os << step.GetOperand();
		break;
	case kAdd:
		os << "+";
		break;
	case kSubtract:
		os << "-";
		break;
	case kMultiply:
		os << "*";
		break;
	case kDivide:
		os << "/";
		break;
	}
	return os;
}




int main(int argc, char **argv)
{
	cout << "Welcome to Eval!\n";

	const int kMaxSteps = 32;
	const int kStackSize = 64;

	string line;
	RPNStep steps[kMaxSteps];
	int nsteps, i;
	double bottom[kStackSize], *stack=bottom;
	bool good;

	for(;;)
	{
		nsteps = 0; // fresh start each time.

		// let them enter it
		for(;;)
		{
			cout << ": "; // our prompt
			cin >> line;

			// NOTE: Should strip leading and trailing whitespace,
			// and actually check that 'number's are really numbers.
			
			if( line == string("q") )
				return 0; // quit application
			else if( line == string("e") ) // done entering
				break; // exit loop
			else if( line == string("+") )
				steps[nsteps++] = RPNStep(kAdd);
			else if( line == string("-") )
				steps[nsteps++] = RPNStep(kSubtract);
			else if( line == string("*") )
				steps[nsteps++] = RPNStep(kMultiply);
			else if( line == string("/") )
				steps[nsteps++] = RPNStep(kDivide);
			else // assume a number
				steps[nsteps++] = RPNStep(kOperand,atof(line.c_str()));
		}

		// print it out
		for(i=0;i<nsteps;i++)
			cout << steps[i] << " ";
		cout << "= ";

		// evaluate it and print out the value
		good = true;
		for(i=0;i<nsteps;i++)
		{
			if( !steps[i].Execute(bottom,&stack) )
			{
				good = false;
				break;
			}
		}
		if( !good || stack <= bottom )
			cout << "Error\n";
		else
			cout << *(--stack) << "\n";
	}

	return 0;
}

